//************************************************************************
//
//	Most of this code is from the Subprocess object.  
//	Numerous additions, deletions, changes and any bugs courtesy of:
//
//		Felipe A. Rodriguez		
//
//	Portions of this file were derived from:
//	
//			Subprocess.m	(v10)
//			by Charles L. Oei
//			pty support by Joe Freeman
//			Subprocess Example, Release 2.0
//			NeXT Computer, Inc. 
//
//			pppkill.c
//			by Scott Hess
//
//	This code is supplied "as is" the author makes no warranty as to its 
//	suitability for any purpose.  This code is free and may be distributed 
//	in accordance with the terms of the:
//		
//			GNU GENERAL PUBLIC LICENSE
//			Version 2, June 1991
//			copyright (C) 1989, 1991 Free Software Foundation, Inc.
// 			675 Mass Ave, Cambridge, MA 02139, USA
//
//************************************************************************
#import "Subprocess.h"
#import "Coordinator.h"

#import <appkit/nextstd.h>
#import <appkit/Application.h>
#import <appkit/Panel.h>
#import <ansi/string.h>
#import <dpsclient/dpsNeXT.h>
#import <defaults/defaults.h>

#define	PTY_TEMPLATE "/dev/pty??"
#define	PTY_LENGTH 11

static void showError();

//************************************************************************
//
// Private Instance Methods
//
//************************************************************************

@interface Subprocess(Private)

- childDidExit;
- fdHandler:(int)theFd;

@end
@implementation Subprocess(Private)

- childDidExit
    // cleanup after a child process exits
{
    if (childPid)
		{
		DPSRemoveFD(fromChild);
		close(fromChild);
		fclose(fpToChild);
		childPid=0;				// specify that child is dead
		if (delegate && [delegate respondsTo:@selector(subprocessDone)])
			[delegate perform:@selector(subprocessDone)];
		}
		
    return self;
}
//************************************************************************
//
//		Called when output from UNIX subprocess is available	
//
//************************************************************************

- fdHandler:(int)theFd
    // DPS handler for output from subprocess
{
    if (((bufferCount = read(theFd, outputBuffer, BUFFERSIZE-1)) < 0) || 	
			(!bufferCount))
		{
		[self childDidExit];
		return self;
		}
    outputBuffer[bufferCount] = '\0';
    if (delegate && [delegate respondsTo:@selector(subprocessOutput:)])
		[delegate perform:@selector(subprocessOutput:)                	
			with:(void *)&outputBuffer];
			
    return self;
}

@end

//************************************************************************
//
//		Private Utility Routines
//
//************************************************************************
 
static void showError (const char *errorString, id theDelegate)
    // ensure errors never get dropped on the floor
{
    if(theDelegate && [theDelegate respondsTo:@selector(subprocessError:)])
		[theDelegate perform:@selector(subprocessError:)
	    	with:(void *)errorString];
    else if(NXApp)		// no delegate, but we're running w/in an App
			NXRunAlertPanel(0, errorString, 0, 0, 0);
    	 else
			perror(errorString); 				//output errstr on stderr
}
//*****************************************************************************
//
//		DPS registered function which recieves output from UNIX subprocess
//
//		void DPSAddFD(int fd, DPSFDProc handler, void *userData, int priority)
//      fdhandler registered to read file descriptor fromChild
//
//*****************************************************************************

static void fdHandler(int theFd, id self)
{
    [self fdHandler:theFd];
}
//*****************************************************************************
//
//		attempt to setup the pty pair
//
//*****************************************************************************

static void getptys (int *master, int *slave)
{
    char device[PTY_LENGTH];
    char *block, *num;
    char *blockLoc; // specifies the location of block for the device string
    char *numLoc; 	// specifies the pty name with the digit ptyxD
    char *msLoc;	// specifies the master (ptyxx) or slave (ttyxx)
    struct sgttyb setp = {B9600, B9600, (char)0x7f, (char)0x15, (CRMOD|ANYP)};
    struct tchars setc = {CINTR, CQUIT, CSTART, CSTOP, CEOF, CBRK};
    struct ltchars sltc = {CSUSP, CDSUSP, CRPRNT, CFLUSH, CWERASE, CLNEXT};
    int	lset = (LCRTBS|LCRTERA|LCRTKIL|LCTLECH|LPENDIN|LDECCTQ);
    int	setd = NTTYDISC;		/* new tty discipline */
    
    strcpy(device, PTY_TEMPLATE); 	// string constants are not writable
    blockLoc = &device[ strlen("/dev/pty") ];
    numLoc = &device[ strlen("/dev/pty?") ];
    msLoc = &device[ strlen("/dev/") ];
    for (block = "pqrs"; *block; block++)	
		{									
		*blockLoc = *block;
		for (num = "0123456789abcdef"; *num; num++)
			{
			*numLoc = *num;
			*master = open(device, O_RDWR);	//Open device assign fd to master
			if (*master >= 0)				//If master 0 or >, open succesful
				{
				*msLoc = 't';
				*slave = open(device, O_RDWR);
				if (*slave >= 0)
					{
						/* set parameters -- stty */
					(void) ioctl(*slave, TIOCSETP, (char *)&setp);
						/* set special characters */
					(void) ioctl(*slave, TIOCSETC, (char *)&setc);
						/* set line discipline */
					(void) ioctl(*slave, TIOCSETD, (char *)&setd);
						/* set local special chars */
					(void) ioctl(*slave, TIOCSLTC, (char *)&sltc);
						/* set entire local mode word */
					(void) ioctl(*slave, TIOCLSET, (char *)&lset);
					
					return;
					}
				}
			} /* hunting through a bank of ptys */
		} /* hunting through blocks of ptys in all the right places */
		
    *master = -1;		//If we get here we were unable to open 
    *slave = -1;		//master and/or slave pty's 
}

@implementation Subprocess

//************************ Public Instance Methods ****************************
//*****************************************************************************
//
// 		initializes an instance of Subprocess and corresponding UNIX process
//
//*****************************************************************************

- init:(const char *)subprocessString withDelegate:theDelegate
    		andStdErr:(BOOL)wantsStdErr
{
static char *argList[32], delim[] = {" "}, *argString;
static int nread, pidChild2, n = 0;
static char buff[BUFFERSIZE]; 
static char logFileName[MAXPATHLEN + 1];		// syslogd logging file
int tty, numFds, fd;							// for temporary use
int processGroup;
int pidChild;			// needed because childPid does not exist
						// until Subprocess is instantiated
	if(theDelegate && [theDelegate respondsTo:@selector(syslogdReset)])
		[theDelegate perform:@selector(syslogdReset)];	// reset syslogd daemon
    								// /dev/tty kernel synm for controlling tty
	tty = open("/dev/tty", O_RDWR);			// open the controlling tty 
	getptys(&masterPty,&slavePty);			// find and open psuedo terms
	strcpy(logFileName, NXGetDefaultValue([NXApp appName], "locFIFO"));
	if (masterPty <= 0 || slavePty <= 0)
		{
		showError("Error grabbing ptys for subprocess.", theDelegate);
		return self;
		}
			// remove the controlling tty if launched from a shell,
			// but not Workspace; so that we have job control over
			// the parent application in shell and so that subprocesses 
			// can be restarted in Workspace
	if((tty<0) && ((tty = open("/dev/tty", 2))>=0))
		{
		ioctl(tty, TIOCNOTTY, 0);
		close(tty);
		}
    switch (pidChild = fork())		// create new process without fully copying
   	 	{							// the address space of the old process
    	case -1:	// error
		
			showError("Error starting UNIX vfork of subprocess.", theDelegate);
			return self;
			
    	case 0:  	// child -- Vfork returns 0 in the child's context
		
			dup2(slavePty, 0);		// dup slavePty to be stdin
			dup2(slavePty, 1);		// dup slavePty to be stdout
			if(wantsStdErr)
				dup2(slavePty, 2);	// dup slavePty to be stderr
			numFds = getdtablesize();
			for(fd=3; fd < numFds; fd++)		// Child should close all open
				close(fd);						// fd's beyond our basic 3
			processGroup = getpid();			// return current pid
				// think -- tty was opened by parent so set pgrp of tty
				//	to that of ourselves
			ioctl(0, TIOCSPGRP, (char *)&processGroup);		
			setpgrp (0, processGroup);  //make curr process process grp leader
			if((log = open(logFileName, O_RDONLY)) < 0)		// open FIFO
				perror("Unable to open FIFO special file.");

			switch (pidChild2 = vfork())				
				{							
				case -1:	// error
				
					showError("Error during vfork to exec pppd", theDelegate);
					return self;
					
				case 0:  	// child -- Vfork returns 0 in the child's context

					argString = NXCopyStringBuffer(subprocessString);
					argList[0] = strtok(argString, delim);	// Parse commLine
					while(argList[n++] != NULL)				// to an array of
						{									// strings
						argList[n] = strtok(NULL, delim); 
						}
					execv(NXGetDefaultValue([NXApp appName], "pppdPath"), 	
																	argList);
					perror("error exec'ing pppd"); 
					exit(1);
		
				default:	// parent --  vfork returns pid of the child  		
						
					for(;;)		// reads log and writes to pty/user 
						{
						if((nread = read(log, buff, BUFFERSIZE)) < 0)
							{
							perror("Error reading FIFO.");
							break;
							}
						if(write(1, buff, nread) != nread)
							perror("Error writing to Slave Pty.");
						}
					wait(NULL);		// siesta time
				}		
			perror("vfork 1st (child)"); 
			exit(1);
			
    	default:	// parent --  vfork returns pid of the child in the 		
					//				parent's context
			[self setDelegate:theDelegate];
			childPid = pidChild;
			close(slavePty);
			fpToChild = fdopen(masterPty, "w");	//return ptr to masterPty
			fromChild = masterPty;
			setbuf(fpToChild, NULL);	// no buffering
				// Registers fdhandler to be called when their is activity with 
				// file descriptor fromChild
				// void DPSAddFD(int fd, DPSFDProc handler, void *userData, 
				// int priority)    
			DPSAddFD(fromChild, (DPSFDProc)fdHandler, (id)self, 	
					NX_BASETHRESHOLD+1);

			return self;
    	}
}

//*****************************************************************************
//
// 		Attempt to kill our child as well as pppd and its subprocesses at any // 		stage in their respective lifetimes.  This is trickier than it sounds.
//
//*****************************************************************************

- terminate:sender
{
FILE *ff; 
int pid;
int prgrp;
char lockPath[32] = {"rm usr/spool/uucp/LCK/LCK.."};	
					
    if((ff = fopen( "/etc/ppp/ppp0.pid", "r")) == NULL)
		{
		if(childPid)
			{								// get the child's process group
			prgrp =  getpgrp(childPid);		// kill the child's group with
			killpg(prgrp, SIGKILL);			// extreme prejudice
			strcpy(Path, lockPath);		
			strcat(Path, NXGetDefaultValue([NXApp appName], "modemPort"));				
			system(Path);					// remove uucp lock from port
			}
		}
	else			// attempt to do a clean hit on a properly running pppd 
		{
    	if( fscanf( ff, "%d", &pid)<1) 
        	perror("Unable to read pid from ppp0.pid\n");
    	fclose( ff);
    	if( kill( pid, SIGINT)==-1) 
        	perror("killing pppd");
		if (childPid)
			kill(childPid, SIGTERM);
    	}
	[self childDidExit];
	
    return self;
}

- setDelegate:anObject
{
    delegate = anObject;
	
    return self;
}

- delegate
{
    return delegate;
}

@end

