/*
 * options.c - handles option processing for PPP.
 *
 * Copyright (c) 1989 Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Carnegie Mellon University.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *	Modifications added 6/16/96 to support GateKeeper.app
 *	Copyright (c) 1996
 *		
 *			by	Felipe A. Rodriguez
 *
 *	GateKeeper code is supplied "as is" the author makes no warranty as  
 *	to its suitability for any purpose.  This code is free and may be 
 *	distributed in accordance with the terms of the:
 *		
 *			GNU GENERAL PUBLIC LICENSE
 *			Version 2, June 1991
 *			copyright (C) 1989, 1991 Free Software Foundation, Inc.
 * 			675 Mass Ave, Cambridge, MA 02139, USA
 *
 */


//#ifndef lint
//char rcsid[] = "$Id: options.c,v 1.22 1995/06/30 01:52:24 paulus Exp $";
//#endif

#import <objc/Object.h>
#import <appkit/Application.h>
#import <appkit/nextstd.h>
#import "options.h"
#import "Coordinator.h"
#import "OptionsEditor.h"


#define	S_IRUSR	_S_IRUSR		/* read permission, */
#define NGROUPS_MAX	16     		/* Number of simultaneous    */  
	                       		/* supplementary group IDs   */
			      				/* per process.              */
#define S_IRGRP	0000040		/* read permission, group */
#define S_IROTH	0000004		/* read permission, other */
#define MAXWORDLEN	1024	/* max length of word in file (incl null) */
#define MAXARGS		1		/* max # args to a command */
#define MAXNAMELEN	256		/* max length of hostname or name for auth */
#define MAXSECRETLEN	256	/* max length of password or secret */

#import <stdio.h>
#import <errno.h>
#import <ctype.h>
#import <unistd.h>
#import <limits.h>
#import <stdlib.h>
#import <termios.h>
#import <syslog.h>
#import <string.h>
#import <netdb.h>
#import <pwd.h>
#import <sys/types.h>
#import <sys/stat.h>
#import <sys/param.h>		/* for MAXPATHLEN and BSD4_4, if defined */
#import <netinet/in.h>


typedef unsigned u_int32_t;

#ifndef GIDSET_TYPE
#define GIDSET_TYPE	int
#endif

/*
 * Option variables and default values.
 */
int	debug = 0;		/* Debug flag */
int	default_device = 1;	/* Using /dev/tty or equivalent */
char devnam[MAXPATHLEN] = "/dev/tty";	/* Device name */
int	crtscts = 0;		/* Use hardware flow control */
int	modem = 1;		/* Use modem control lines */
int	inspeed = 0;		/* Input/Output speed requested */
int	lockflag = 0;		/* Create lock file to lock the serial dev */
int	nodetach = 0;		/* Don't detach from controlling tty */
int	defaultroute = 0;	/* assign default route through interface */
int	proxyarp = 0;		/* Set up proxy ARP entry for peer */
    FILE *f;


/*
 * Prototypes
 */
static int setspeed (char *arg);
static int setlock (void);
static int clearlock (void);
static int setdefaultroute (void);
static int cleardefaultroute (void);
static int setdebug (void);
static int cleardebug (void);
static int setd (void);
static int cleard (void);
static int setcrtscts (void);
static int clearcrtscts (void);
static int setdevname(char *cp);

static int hitArray[7];		// used in determining whether an option flag
							// method was called while scanning options file 


/*
 * Valid arguments.
 */
static struct cmd {
    char *cmd_name;
    int num_args;
    int (*cmd_func)();
} cmds[] = {
    {"-d", 0, setd},	/* Increase debugging level */
    {"@-d", 0, cleard},	/* Increase debugging level */
    {"debug", 0, setdebug},	/* Increase debugging level */
    {"@debug", 0, cleardebug},	/* Increase debugging level */
    {"crtscts", 0, setcrtscts},		/* set h/w flow control */
    {"-crtscts", 0, clearcrtscts}, 	/* clear h/w flow control */
    {"@crtscts", 0, clearcrtscts}, 	/* clear h/w flow control */
    {"lock", 0, setlock},	/* Lock serial device (with lock file) */
    {"@lock", 0, clearlock},	/* Lock serial device (with lock file) */
    {"defaultroute", 0, setdefaultroute}, /* Add default route */
    {"@defaultroute", 0, cleardefaultroute}, /* Add default route */
    {NULL, 0, NULL}
};


/*
 * Read a word from a file.
 * Words are delimited by white-space or by quotes (" or ').
 * Quotes, white-space and \ may be escaped with \.
 * \<newline> is ignored.
 */

int
getword(f, word, newlinep, filename)
    FILE *f;
    char *word;
    int *newlinep;
    const char *filename;
{
    int c, len, escape;
    int quoted, comment;
    int value, digit, got, n;

#define isoctal(c) ((c) >= '0' && (c) < '8')

    *newlinep = 0;
    len = 0;
    escape = 0;
    comment = 0;

    /*
     * First skip white-space and comments.
     */
    for (;;) {
	c = getc(f);
	if (c == EOF)
	    break;

	/*
	 * A newline means the end of a comment; backslash-newline
	 * is ignored.  Note that we cannot have escape && comment.
	 */
	if (c == '\n') {
	    if (!escape) {
		*newlinep = 1;
		comment = 0;
	    } else
		escape = 0;
	    continue;
	}


	/*
	 * Possible one of our flags or keywords.
	 */
	if (comment && (c == '@'))
	    break;

	/*
	 * Ignore characters other than newline in a comment.
	 */
	if (comment)
	    continue;


	/*
	 * If this character is escaped, we have a word start.
	 */
	if (escape)
	    break;

	/*
	 * If this is the escape character, look at the next character.
	 */
	if (c == '\\') {
	    escape = 1;
	    continue;
	}

	/*
	 * If this is the start of a comment, ignore the rest of the line.
	 */
	if (c == '#') {
	    comment = 1;
	    continue;
	}

	/*
	 * A non-whitespace character is the start of a word.
	 */
	if (!isspace(c))
	    break;
    }

    /*
     * Save the delimiter for quoted strings.
     */
    if (!escape && (c == '"' || c == '\'')) {
        quoted = c;
	c = getc(f);
    } else
        quoted = 0;

    /*
     * Process characters until the end of the word.
     */
    while (c != EOF) {
	if (escape) {
	    /*
	     * This character is escaped: backslash-newline is ignored,
	     * various other characters indicate particular values
	     * as for C backslash-escapes.
	     */
	    escape = 0;
	    if (c == '\n') {
	        c = getc(f);
		continue;
	    }

	    got = 0;
	    switch (c) {
	    case 'a':
		value = '\a';
		break;
	    case 'b':
		value = '\b';
		break;
	    case 'f':
		value = '\f';
		break;
	    case 'n':
		value = '\n';
		break;
	    case 'r':
		value = '\r';
		break;
	    case 's':
		value = ' ';
		break;
	    case 't':
		value = '\t';
		break;

	    default:
		if (isoctal(c)) {
		    /*
		     * \ddd octal sequence
		     */
		    value = 0;
		    for (n = 0; n < 3 && isoctal(c); ++n) {
			value = (value << 3) + (c & 07);
			c = getc(f);
		    }
		    got = 1;
		    break;
		}

		if (c == 'x') {
		    /*
		     * \x<hex_string> sequence
		     */
		    value = 0;
		    c = getc(f);
		    for (n = 0; n < 2 && isxdigit(c); ++n) {
			digit = toupper(c) - '0';
			if (digit > 10)
			    digit += '0' + 10 - 'A';
			value = (value << 4) + digit;
			c = getc (f);
		    }
		    got = 1;
		    break;
		}

		/*
		 * Otherwise the character stands for itself.
		 */
		value = c;
		break;
	    }

	    /*
	     * Store the resulting character for the escape sequence.
	     */
	    if (len < MAXWORDLEN-1)
		word[len] = value;
	    ++len;

	    if (!got)
		c = getc(f);
	    continue;

	}

	/*
	 * Not escaped: see if we've reached the end of the word.
	 */
	if (quoted) {
	    if (c == quoted)
		break;
	} else {
	    if (isspace(c) || c == '#') {
		ungetc (c, f);
		break;
	    }
	}

	/*
	 * Backslash starts an escape sequence.
	 */
	if (c == '\\') {
	    escape = 1;
	    c = getc(f);
	    continue;
	}

	/*
	 * An ordinary character: store it in the word and get another.
	 */
	if (len < MAXWORDLEN-1)
	    word[len] = c;
	++len;

	c = getc(f);
    }

    /*
     * End of the word: check for errors.
     */
    if (c == EOF) {
	if (ferror(f)) {
	    if (errno == 0)
		errno = EIO;
	    perror(filename);
	    return 0;
	}
	/*
	 * If len is zero, then we didn't find a word before the
	 * end of the file.
	 */
	if (len == 0)
	    return 0;
    }

    /*
     * Warn if the word was too long, and append a terminating null.
     */
    if (len >= MAXWORDLEN) {
	perror("warning: word in options file is too long");
	len = MAXWORDLEN - 1;
    }
    word[len] = 0;

    return 1;

#undef isoctal

}

/*
 * options_from_file - Read a string of options from a file,
 * and interpret them.
 */
int
options_from_file(filename, must_exist, check_prot)
    const char *filename;
    int must_exist;
    int check_prot;
{
    int i, newline, ret;
    struct cmd *cmdp;
    char *argv[MAXARGS];
    char args[MAXARGS][MAXWORDLEN];
    char cmd[MAXWORDLEN];

	for (i = 0; i < 6; i++) 		// zero out hit array each invocation
 		{hitArray[i] = 0;}		// used in detecting whether func was called

	if ((f = fopen(filename, "r+")) == NULL) {
	if (!must_exist && errno == ENOENT)
	    return 1;
	perror(filename);
	return 0;
    }
    if (check_prot && !readable(fileno(f))) {
	fprintf(stderr, "%s: access denied\n", filename);
	fclose(f);
	return 0;
    }

    while (getword(f, cmd, &newline, filename)) {
	/*
	 * First see if it's a command.
	 */
	for (cmdp = cmds; cmdp->cmd_name; cmdp++)
	    if (!strcmp(cmd, cmdp->cmd_name))
		break;

	if (cmdp->cmd_name != NULL) {
	    for (i = 0; i < cmdp->num_args; ++i) {
		if (!getword(f, args[i], &newline, filename)) {
		    fprintf(stderr,
			    "In file %s: too few parameters for command %s\n",
			    filename, cmd);
		    fclose(f);
			f = NULL;
		    return 0;
		}
		argv[i] = args[i];
	    }
	    if (!(*cmdp->cmd_func)(argv)) {
		fclose(f);
		f = NULL;
		return 0;
	    }

	} else {
	    /*
	     * Maybe a tty name, speed or IP address?
	     */
	    ret = setdevname(cmd);
		ret = setspeed(cmd);
		}
	}
	if(!hitArray[4])
		{
		fputs("\n/dev/cua  \n", f);
		[oDelegate setDevName:"/dev/cua"];
		}
	if(!hitArray[5])
		{
		fputs("\n57600  \n", f);
		[oDelegate setSpeed:"57600"];
		}
	if(!hitArray[0] && [oDelegate defaultRoute])
		fputs("\ndefaultroute  \n", f);
	if(!hitArray[1] && [oDelegate lockFlag])
		fputs("\nlock  \n", f);
	if(!hitArray[2] && [oDelegate crtscts])
		fputs("\ncrtscts  \n", f);
	if(!hitArray[3] && [oDelegate debug])
		fputs("\ndebug  \n", f);

	if(f)
		fclose(f);
		f = NULL;

    return 1;
}

/*
 * readable - check if a file is readable by the real user.
 */
int
readable(fd)
    int fd;
{
    uid_t uid;
    int ngroups, i;
    struct stat sbuf;
    GIDSET_TYPE groups[NGROUPS_MAX];

    uid = getuid();
    if (uid == 0)
	return 1;
    if (fstat(fd, &sbuf) != 0)
	return 0;
    if (sbuf.st_uid == uid)
	return sbuf.st_mode & S_IRUSR;
    if (sbuf.st_gid == getgid())
	return sbuf.st_mode & S_IRGRP;
    ngroups = getgroups(NGROUPS_MAX, groups);
    for (i = 0; i < ngroups; ++i)
	if (sbuf.st_gid == groups[i])
	    return sbuf.st_mode & S_IRGRP;
    return sbuf.st_mode & S_IROTH;
}
//*****************************************************************************
//
// 		set and remove "lock" flag 
//
//*****************************************************************************

static int
setlock()
{
	if([oDelegate shouldSave] && ![oDelegate lockFlag])
		{
     	fseek(f,-4, SEEK_CUR);
		fputs("#@lock\n", f);
	    fseek(f, ftell(f), SEEK_SET);
		}
	else
		[oDelegate setLockFlag];
 	hitArray[1] = 1;

    return 1;
}
static int
clearlock()
{
	if([oDelegate shouldSave] && [oDelegate lockFlag])
		{
     	fseek(f,-6, SEEK_CUR);
		fputs("lock  \n", f);
	    fseek(f, ftell(f), SEEK_SET);
		}
 	hitArray[1] = 1;

    return 1;
}
//*****************************************************************************
//
// 		set and remove "defaultRoute" flag 
//
//*****************************************************************************

static int
setdefaultroute()
{
	if([oDelegate shouldSave] && ![oDelegate defaultRoute])
		{
     	fseek(f,-12, SEEK_CUR);
		fputs("#@defaultroute\n", f);
	    fseek(f, ftell(f), SEEK_SET);
		}
	else
		[oDelegate setDefaultRoute];
 	hitArray[0] = 1;

    return 1;
}
static int
cleardefaultroute()
{
	if([oDelegate shouldSave] && [oDelegate defaultRoute])
		{
     	fseek(f,-14, SEEK_CUR);
		fputs("defaultroute  \n", f);
	    fseek(f, ftell(f), SEEK_SET);
		}
 	hitArray[0] = 1;

    return 1;
}
//*****************************************************************************
//
// 		set and remove "debug" flag 
//
//*****************************************************************************

static int
setdebug()
{
	if([oDelegate shouldSave] && ![oDelegate debug])
		{
     	fseek(f,-5, SEEK_CUR);
		fputs("#@debug\n", f);
	    fseek(f, ftell(f), SEEK_SET);
		}
	else
		[oDelegate setDebug];
 	hitArray[3] = 1;

    return 1;
}
static int
cleardebug()
{
	if([oDelegate shouldSave] && [oDelegate debug])
		{
     	fseek(f,-7, SEEK_CUR);
		fputs("debug  \n", f);
	    fseek(f, ftell(f), SEEK_SET);
		}
 	hitArray[3] = 1;

    return 1;
}
static int
setd()
{
	if([oDelegate shouldSave] && ![oDelegate debug])
		{
     	fseek(f,-2, SEEK_CUR);
		fputs("#@-d\n", f);
	    fseek(f, ftell(f), SEEK_SET);
		}
	else
		[oDelegate setDebug];
 	hitArray[3] = 1;

    return 1;
}
static int
cleard()
{
	if([oDelegate shouldSave] && [oDelegate debug])
		{
     	fseek(f,-4, SEEK_CUR);
		fputs("-d  \n", f);
	    fseek(f, ftell(f), SEEK_SET);
		}
 	hitArray[3] = 1;

    return 1;
}
//*****************************************************************************
//
// 		set and remove "crtscts" flag 
//
//*****************************************************************************

static int
setcrtscts()
{
	if([oDelegate shouldSave] && ![oDelegate crtscts])
		{
     	fseek(f,-7, SEEK_CUR);
		fputs("-crtscts\n", f);
	    fseek(f, ftell(f), SEEK_SET);
		}
	else
		[oDelegate setCrtscts];
 	hitArray[2] = 1;

    return 1;
}
static int
clearcrtscts()
{
	if([oDelegate shouldSave] && [oDelegate crtscts])
		{
     	fseek(f,-8, SEEK_CUR);
		fputs("crtscts \n", f);
	    fseek(f, ftell(f), SEEK_SET);
		}
 	hitArray[2] = 1;

    return 1;
}

/*
 * setspeed - Set the speed.
 */
static int
setspeed(arg)
    char *arg;
{
    char *ptr;
    int spd;
	int lenOfarg = strlen(arg);

    spd = strtol(arg, &ptr, 0);
    if (ptr == arg || *ptr != 0 || spd == 0)
	return 0;

	if([oDelegate shouldSave] && (strcmp([oDelegate speed], arg) != 0))
		{
     	fseek(f,-lenOfarg, SEEK_CUR);
		fputs([oDelegate speed], f);
		if((lenOfarg = strlen([oDelegate speed])) == 5)
			fputs("\n", f);
		if(lenOfarg == 4)
			fputs(" \n", f);
		if(lenOfarg == 3)
			fputs("  \n", f);
	    fseek(f, ftell(f), SEEK_SET);
 		hitArray[5] = 1;
		}
	else
		{
		if([oDelegate setSpeed:arg])
 			hitArray[5] = 1;
		}

    inspeed = spd;
    return 1;
}


/*
 * setdevname - Set the device name.
 */
int
setdevname(cp)
    char *cp;
{
    struct stat statbuf;
    char *ttyname();
    char dev[MAXPATHLEN];
	int lenOfcp = strlen(cp);

    if (strncmp("/dev/", cp, 5) != 0) {
	strcpy(dev, "/dev/");
	strncat(dev, cp, MAXPATHLEN - 5);
	dev[MAXPATHLEN-1] = 0;
	cp = dev;
    }

    /*
     * Check if there is a device by this name.
     */
    if (stat(cp, &statbuf) < 0) {
	if (errno == ENOENT)
	    return 0;
	syslog(LOG_ERR, cp);
	return -1;
    }
  
	if([oDelegate shouldSave] && (strcmp([oDelegate devName], cp) != 0))
		{
     	fseek(f,-lenOfcp, SEEK_CUR);
		fputs([oDelegate devName], f);
		if((lenOfcp = strlen([oDelegate devName]))== 10)
			fputs("\n", f);
		if(lenOfcp == 9)
			fputs(" \n", f);
		if(lenOfcp == 8)
			fputs("  \n", f);
	    fseek(f, ftell(f), SEEK_SET);
		}
	else
		[oDelegate setDevName:cp];
 	hitArray[4] = 1;

    (void) strncpy(devnam, cp, MAXPATHLEN);
    devnam[MAXPATHLEN-1] = 0;
    default_device = FALSE;
  
    return 1;
}

int setDelegate(id del)
{
	oDelegate = del;

    return 1;
}
